@@ -19,26 +19,38 @@
 
 #include <grpc/support/port_platform.h>
 
+#include <algorithm>
+#include <map>
 #include <memory>
 #include <set>
 #include <string>
 #include <vector>
 
-#include "absl/container/inlined_vector.h"
+#include "absl/status/status.h"
+#include "absl/strings/string_view.h"
 
-#include <grpc/slice.h>
-
-#include "src/core/ext/xds/certificate_provider_store.h"
-#include "src/core/lib/gprpp/memory.h"
-#include "src/core/lib/gprpp/ref_counted_ptr.h"
 #include "src/core/lib/iomgr/error.h"
 #include "src/core/lib/json/json.h"
-#include "src/core/lib/security/credentials/credentials.h"
 
 namespace grpc_core {
 
+bool XdsFederationEnabled();
+
 class XdsClient;
 
+class XdsCertificateProviderPluginMapInterface {
+ public:
+  virtual ~XdsCertificateProviderPluginMapInterface() = default;
+
+  virtual absl::Status AddPlugin(const std::string& instance_name,
+                                 const std::string& plugin_name,
+                                 const Json& config) = 0;
+
+  virtual bool HasPlugin(const std::string& instance_name) const = 0;
+
+  virtual std::string ToString() const = 0;
+};
+
 class XdsBootstrap {
  public:
   struct Node {
@@ -78,21 +90,28 @@ class XdsBootstrap {
     Json::Object ToJson() const;
 
     bool ShouldUseV3() const;
+    bool IgnoreResourceDeletion() const;
   };
 
   struct Authority {
     std::string client_listener_resource_name_template;
-    absl::InlinedVector<XdsServer, 1> xds_servers;
+    std::vector<XdsServer> xds_servers;
   };
 
   // Creates bootstrap object from json_string.
   // If *error is not GRPC_ERROR_NONE after returning, then there was an
   // error parsing the contents.
-  static std::unique_ptr<XdsBootstrap> Create(absl::string_view json_string,
-                                              grpc_error_handle* error);
+  static std::unique_ptr<XdsBootstrap> Create(
+      absl::string_view json_string,
+      std::unique_ptr<XdsCertificateProviderPluginMapInterface>
+          certificate_provider_plugin_map,
+      grpc_error_handle* error);
 
   // Do not instantiate directly -- use Create() above instead.
-  XdsBootstrap(Json json, grpc_error_handle* error);
+  XdsBootstrap(Json json,
+               std::unique_ptr<XdsCertificateProviderPluginMapInterface>
+                   certificate_provider_plugin_map,
+               grpc_error_handle* error);
 
   std::string ToString() const;
 
@@ -110,16 +129,16 @@ class XdsBootstrap {
     return authorities_;
   }
   const Authority* LookupAuthority(const std::string& name) const;
-  const CertificateProviderStore::PluginDefinitionMap& certificate_providers()
-      const {
-    return certificate_providers_;
+  const XdsCertificateProviderPluginMapInterface*
+  certificate_provider_plugin_map() const {
+    return certificate_provider_plugin_map_.get();
   }
   // A util method to check that an xds server exists in this bootstrap file.
   bool XdsServerExists(const XdsServer& server) const;
 
  private:
-  grpc_error_handle ParseXdsServerList(
-      Json* json, absl::InlinedVector<XdsServer, 1>* servers);
+  grpc_error_handle ParseXdsServerList(Json* json,
+                                       std::vector<XdsServer>* servers);
   grpc_error_handle ParseAuthorities(Json* json);
   grpc_error_handle ParseAuthority(Json* json, const std::string& name);
   grpc_error_handle ParseNode(Json* json);
@@ -128,14 +147,15 @@ class XdsBootstrap {
   grpc_error_handle ParseCertificateProvider(const std::string& instance_name,
                                              Json* certificate_provider_json);
 
-  absl::InlinedVector<XdsServer, 1> servers_;
+  std::vector<XdsServer> servers_;
   std::unique_ptr<Node> node_;
   std::string client_default_listener_resource_name_template_;
   std::string server_listener_resource_name_template_;
   std::map<std::string, Authority> authorities_;
-  CertificateProviderStore::PluginDefinitionMap certificate_providers_;
+  std::unique_ptr<XdsCertificateProviderPluginMapInterface>
+      certificate_provider_plugin_map_;
 };
 
 }  // namespace grpc_core
 
-#endif /* GRPC_CORE_EXT_XDS_XDS_BOOTSTRAP_H */
+#endif  // GRPC_CORE_EXT_XDS_XDS_BOOTSTRAP_H
