@@ -21,7 +21,10 @@
 
 #include <grpc/support/port_platform.h>
 
+#include "src/core/lib/channel/channel_fwd.h"
+#include "src/core/lib/channel/channel_stack.h"
 #include "src/core/lib/channel/channel_stack_builder.h"
+#include "src/core/lib/transport/transport.h"
 
 extern const grpc_channel_filter grpc_connected_filter;
 
